/*  GSML - General Sound Manipulation Library
 *  Copyright (C) 2000-2001 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Grouping of several TrackLayer to one "undo-step", multi-track
 *   data storage.
 */

#include "DataLayer.hh"

#include <algorithm>
#include <functional>

using namespace std;
using namespace GSML;

/*************************
 * The DataLayer::Marker *
 *************************/

DataLayer::Markers::Marker::Marker ()
  : name (""), position (), undeleteable (false)
{
}

DataLayer::Markers::Marker::Marker (const DataLayer::Markers::Marker& source)
{
  operator= (source);
}

DataLayer::Markers::Marker& DataLayer::Markers::Marker::operator= (const DataLayer::Markers::Marker& source)
{
  if (this != &source)
    {
      name = source.name;
      position = source.position;
    }
  return *this;
}

DataLayer::Markers::Markers (DataLayer* i_layer)
{
  m_layer = i_layer;
}
	  
DataLayer::Markers::Markers (const Markers& source)
{
  operator= (source);
}

DataLayer::Markers& DataLayer::Markers::operator= (const DataLayer::Markers& source)
{
  if (this != &source)
    {
      markers = source.markers;
      m_layer = source.m_layer;
    }
}

void DataLayer::Markers::SetLayer (DataLayer* n_layer)
{
  m_layer = n_layer;
}

bool DataLayer::Markers::AddMarker (const string& name, sample_index pos, bool undeleteable)
{
  if (m_layer->fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data layer is fiaxted";
      throw e;
    }
	    
  Marker m;
  m.name = name;
  m.position = pos;
	    
  return AddMarker (m);
}
	  
bool DataLayer::Markers::AddMarker (DataLayer::Markers::Marker& new_marker)
{
  if (m_layer->fixated)
    {
      ExceptFixated e;
      e.info = "the underlying data layer is fiaxted";
      throw e;
    }
  
  for (iterator i = markers.begin (); i != markers.end (); ++i)
    if (i->name == new_marker.name)
      return false;
	    
  // sorted inserting # nearly the same code is in SetMarker - REMEMBER to chage it, too #
  for (iterator i = markers.begin (); i != markers.end (); ++i)
    if (i->position > new_marker.position)
      {
	markers.insert (i, new_marker);
	SignalAdded.emit (new_marker.name, new_marker.position);
	return true;
      }
	    
  markers.push_back (new_marker);
  SignalAdded.emit (new_marker.name, new_marker.position);
  return true;
}

bool DataLayer::Markers::DeleteMarker (const string& name)
{
  if (m_layer->fixated)
    {
      ExceptFixated e;
      e.info = "The underlying data layer is fiaxted.";
      throw e;
    }
  
  for (iterator i = markers.begin (); i != markers.end (); ++i)
    {
      if (i->name == name)
	{
	  if (i->undeleteable)
	    {
	      ExceptUndeleteable e;
	      e.info = "The marker is marked as undeleteble.";
	      throw e;
	    }
	  sample_index t_pos = i->position;
	  markers.erase (i);
	  SignalDeleted.emit (name, t_pos);
	  
	  return true;
	}
    }
  return false;
}
	  
bool DataLayer::Markers::GetMarker (const string& name, sample_index* pos)
{
  for (iterator i = markers.begin (); i != markers.end (); ++i)
    {
      if (i->name == name)
	{
	  *pos = i->position;
	  return true;
	}
    }
  return false;
}

bool DataLayer::Markers::SetMarker (const string& name, sample_index pos)
{
  sample_index t_old_pos;
  
  bool rEnd = false;
  for (iterator i = markers.begin (); !rEnd && i != markers.end (); ++i)
    {
      if (i->name == name)
	{
	  t_old_pos = i->position;
	  markers.erase (i);
	  rEnd = true;
	}
    }
  if (!rEnd)
    return false;
  
  Marker new_marker;
  new_marker.name = name;
  new_marker.position = pos;
  
  // sorted insert # nearly the same code is in AddMarker - REMEMBER to chage it, too #
  for (iterator i = markers.begin (); i != markers.end (); ++i)
    if (i->position > pos)
      {
	markers.insert (i, new_marker);
	SignalChanged.emit (name, t_old_pos, pos);
	return true;
      }
  
  markers.push_back (new_marker);
  SignalChanged.emit (name, t_old_pos, pos);
  return true;

}
	  
DataLayer::Markers::const_iterator DataLayer::Markers::begin ()
{
  return markers.begin ();
}
	  
DataLayer::Markers::const_iterator DataLayer::Markers::end ()
{
  return markers.end ();
}
	  
void DataLayer::Markers::Debug ()
{
  cout << "Markers::Debug:" << endl;
  for (iterator i = markers.begin (); i != markers.end (); i++)
    cout << "  : " << i->name << ", " << i->position << ", undeleteable: " << i->undeleteable << endl;
}

/*****************
 * The DataLayer *
 *****************/

DataLayer::DataLayer (const string& i_comment, DataFileIndex* i_index,
		      PCMDatatype* i_temp_type, sample_rate i_rate)
  : markers (this)
{
  m_comment = i_comment;
  m_fileindex = i_index;
  m_temp_sampletype = i_temp_type;
  m_temp_samplerate = i_rate;

  fixated = false;
}

DataLayer::DataLayer (const DataLayer& source_layer)
  : markers (this)
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::DataLayer (const DataLayer& source_layer)" << endl;
#endif
  
  // for the overloaded = operator !!! - we MUST have vaild inits !!!
  fixated = false;
  
  this->operator= (source_layer);
}


DataLayer::~DataLayer ()
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::~DataLayer: " << endl;
#endif
  // destroy all tracks ...
  del_helper<TrackLayer*> p;
  for_each (m_data.begin (), m_data.end (), p);
}

DataLayer::iterator DataLayer::Begin ()
{
  return m_data.begin ();
}

DataLayer::iterator DataLayer::End ()
{
  return m_data.end ();
}

track_index DataLayer::Tracks () const
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::Tracks:" << endl;
#endif
  return m_data.size ();
}

GSML::TrackLayer* DataLayer::CreateTrack (track_index pos)
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::CreateTrack:" << endl;
#endif
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  if (pos > m_data.size () )
    {
      DataLayer::ExceptLogicalError e;
      e.info = "Track behind last track + 1 ...";
      throw e;
    }
  
  TrackLayer* t_layer = new TrackLayer (m_fileindex, m_temp_sampletype, m_temp_samplerate);
  
  m_data.insert (m_data.begin () + pos, t_layer);
  
  // emit Signal
  SignalTrackCreated.emit (pos);
  
  return t_layer;
  
}

DataLayer::iterator DataLayer::CreateTrack (const DataLayer::iterator& pos)
{
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  track_index t_track = pos - m_data.begin ();
  
  TrackLayer* t_layer = new TrackLayer (m_fileindex, m_temp_sampletype, m_temp_samplerate);
  iterator t_return =  m_data.insert (pos, t_layer);
  
  // emit Signal
  SignalTrackCreated.emit (t_track);
  
  return t_return;
}

void DataLayer::CreateNEmptyTracks (track_index count)
{
  while (Tracks () < count)
    CreateTrack (Tracks () );
  
  while (Tracks () > count)
    EraseTrack (Tracks () - 1);
  
  for (iterator i = m_data.begin (); i != m_data.end (); ++i)
    {
      (*i)->RemoveRegion (0, (*i)->Length () );
    }
}

void DataLayer::EraseTrack (track_index pos)
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::EraseTrack:" << endl;
#endif
 if (fixated)
   {
     DataLayer::ExceptFixated e;
     e.info = "Already fixated! ... ";
     throw e;
   }

 // does this track exist ??
 if (pos >= m_data.size () )
   {
     DataLayer::ExceptLogicalError e;
     e.info = "Track behind last track ...";
     throw e;
   }
 
 delete m_data [pos];
 m_data.erase (m_data.begin () + pos);
 
 // emit Signal
 SignalTrackErased.emit (pos);
}

DataLayer::iterator DataLayer::EraseTrack (const iterator& pos)
{
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  track_index t_track = pos - m_data.begin ();
  
  delete *pos;
  iterator t_return = m_data.erase (pos);
  
  // emit Signal
  SignalTrackErased.emit (t_track);
  
  return t_return;
}

void DataLayer::EraseAllTracks ()
{
  int old_track_count;
  
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::EraseAllTrack:" << endl;
#endif
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  // destroy all tracks ...
  for (iterator i = m_data.begin (); i != m_data.end (); ++ i)
    {
      del_helper2 (*i);
      m_data.pop_front ();
      
      SignalTrackErased.emit (0);
    }
}

const string& DataLayer::Comment () const
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::Comment:" << endl;
#endif
  return m_comment;
}

void DataLayer::Comment (const string& n_comment)
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::Comment:" << endl;
#endif
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  m_comment = n_comment;
}

sample_index DataLayer::Length () const
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::Length:" << endl;
#endif
  sample_index t_length = 0;
  
  for (const_iterator cur_track = m_data.begin (); cur_track != m_data.end (); ++ cur_track)
    t_length = limit_min ( (*cur_track)->Length (), t_length);
  
  return t_length;
}

void DataLayer::Fixate ()
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::Fixate:" << endl;
#endif
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  for_each (m_data.begin (), m_data.end (), mem_fun (&TrackLayer::Fixate) );
  
  fixated = true;
}

void DataLayer::UnFixate ()
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::UnFixate:" << endl;
#endif
  if (!fixated)
    {
      DataLayer::ExceptNotFixated e;
      e.info = "Not fixated! ... ";
      throw e;
    }
  
  for_each (m_data.begin (), m_data.end (), mem_fun (&TrackLayer::UnFixate) );
  
  fixated = false;
}

bool DataLayer::IsFixated ()
{
  return fixated;
}

// create and return a new data_layer with the copy of the given region ...
DataLayer* DataLayer::CreateDataLayer (sample_index start, sample_index length) const
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::CreateDataLayer:" << endl;
#endif
  DataLayer* t_layer;
  int tracks;
  
  t_layer = new DataLayer (m_comment, m_fileindex, m_temp_sampletype, m_temp_samplerate);
  
  for (const_iterator cur_track = m_data.begin (); cur_track != m_data.end (); ++cur_track)
    {
      TrackLayer* t_track;
      cout << "_copy_ track: " << "xx" << endl;
      
      t_track = (*cur_track)->CreateTrackLayer (start, length);
      
      t_layer->m_data.push_back (t_track);
    }
  
  return t_layer;
}
  
// inserts the given data_layer at the start position
void DataLayer::InsertDataLayer (sample_index start,
				 const DataLayer& source_layer)
{
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  iterator track;
  const_iterator src_track;
  
  track = m_data.begin ();
  src_track = source_layer.m_data.begin ();
  while (src_track != source_layer.m_data.end () )
    {
      // does the track exist ??
      if (track == m_data.end () )
	{
	  cout << "createing a new track ... " << endl;
	  
	  TrackLayer* t_layer;
	  t_layer = new TrackLayer (m_fileindex, m_temp_sampletype, m_temp_samplerate);
	  m_data.push_back (t_layer);
	}
      else
	{
	  cout << "track exist ... " << endl;
	}
      
      cout << "inserting track: " << endl;
      (*track)->InsertTrackLayer (start, **src_track);
      cout << "insert finished" << endl;
      
      ++ track;
      ++ src_track;
    }
}

void DataLayer::RemoveRegion (sample_index start,
			      sample_index length)
{
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  for (iterator track = m_data.begin (); track != m_data.end (); ++ track)
    {
      (*track)->RemoveRegion (start, length);
    }
}

DataFileIndex* DataLayer::GetDataFileIndex ()
{
  return m_fileindex;
}

PCMDatatype* DataLayer::TempSampletype ()
{
  return m_temp_sampletype;
}

sample_rate DataLayer::TempSamplerate ()
{
  return m_temp_samplerate;
}

void DataLayer::SetTempSampletype (PCMDatatype* n_type)
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::SetTempSampletype: " << endl;
#endif
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  // not tested ...
  for_each (m_data.begin (), m_data.end (), bind2nd (mem_fun(&TrackLayer::SetTempSampletype), n_type) );
  
  m_temp_sampletype = n_type;
}

void DataLayer::SetTempSamplerate (sample_rate n_rate)
{
  #ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::SetTempSamplerate: " << endl;
#endif
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  // not tested ...
  for_each (m_data.begin (), m_data.end (), bind2nd (mem_fun(&TrackLayer::SetTempSamplerate), n_rate) );
  
  m_temp_samplerate = n_rate;
}

// creates an new copy of this
DataLayer& DataLayer::operator= (const DataLayer& source_layer)
{
#ifdef DEBUG_GSML_DATA_LAYER
  cout << "DataLayer::operator=: " << endl;
#endif
  if (fixated)
    {
      DataLayer::ExceptFixated e;
      e.info = "Already fixated! ... ";
      throw e;
    }
  
  if (this != &source_layer)
    {
      // copy config
      m_comment = source_layer.m_comment;
      m_fileindex = source_layer.m_fileindex;
      m_temp_sampletype = source_layer.m_temp_sampletype;
      m_temp_samplerate  = source_layer.m_temp_samplerate;
      
      // destroy all tracks ...
#ifdef DEBUG_GSML_DATA_LAYER
      cout << "Deleteing the tracks: " << endl;
#endif
      
      del_helper<TrackLayer*> p;
      for_each (m_data.begin (), m_data.end (), p);
      m_data.erase (m_data.begin (), m_data.end () );
      
#ifdef DEBUG_GSML_DATA_LAYER
      cout << endl << "Copying the tracks:" << endl;
#endif
      for (const_iterator src_track = source_layer.m_data.begin ();
	   src_track != source_layer.m_data.end (); ++ src_track)
	{
#ifdef DEBUG_GSML_DATA_LAYER
	  cout << "copying: " << "xx" << endl;
#endif
	  TrackLayer* t_track;
	  
	  t_track = new TrackLayer (*(*src_track) );
	  m_data.push_back (t_track);
	} // end for all members
      
      markers = source_layer.markers;
      markers.SetLayer (this);
      
    } // end if this != source
}

//for direct track acces with [] ...
GSML::TrackLayer* DataLayer::operator[] (track_index track) const
{
  if (track >= m_data.size () )
    {
      DataLayer::ExceptLogicalError e;
      e.info = "Track not available! ... ";
      throw e;
    }
  
  return m_data [track];
}

void DataLayer::Debug () const
{
  cout << "DataLayer::Debug:" << endl;
  
  cout << "  Tracks: " << m_data.size () << endl;
  cout << "  Fixated: " << fixated << endl;
  
  cout << "  my_config:" << endl;
  cout << "    comment: " << m_comment << endl;
  cout << "    data_file_index: " << m_fileindex << endl;
  cout << "    temp_data_sampletyp: " << m_temp_sampletype << endl;
  cout << "    temp_data_samplerate: " << m_temp_samplerate << endl;
  
  // not tested ...
  cout << "for_each Debug ..." << endl;
  for_each (m_data.begin (), m_data.end (), mem_fun (&TrackLayer::Debug) );
    
  cout << "End ..." << endl;
}
